/* am930_mac.c: Handles the 802.11 mac functions
*	--------------------------------------------------------------------
*
*   Linux WLAN 
*
*   The contents of this file are subject to the Mozilla Public
*   License Version 1.0 (the "License"); you may not use this file
*   except in compliance with the License. You may obtain a copy of
*   the License at http://www.mozilla.org/MPL/
*
*   Software distributed under the License is distributed on an "AS
*   IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
*   implied. See the License for the specific language governing
*   rights and limitations under the License.
*
*   The initial developer of the original code is Mark S. Mathews
*   <mark@absoval.com>.  Portions created by Mark S. Mathews
*   are Copyright (C) 1998 AbsoluteValue Software, Inc.  All Rights Reserved.
*   
*	--------------------------------------------------------------------
*
*	The author may be reached as mark@absoval.com, or C/O AbsoluteValue
*	Software Inc., P.O. Box 941149, Maitland, FL, 32794-1149
*
*	Thanks to David Hinds, Donald Becker, and the rest of the Linux
*	developers worldwide for making all of this possible.
*
*	--------------------------------------------------------------------
*
*	Functions to perform the 802.11 specific mac functions for the driver.
*	These should eventually be coded as support routines (a-la the Linux
*	ethernet routines), but are here for now. This is partially because
*	the AMD drivers fake ethernet by embedding ethernet frames inside
*	802.11 frames and I want to maintain compatibility. 
*
*	TODO: Need to add 802.2 support, but I don't know the magic #s for
*		802.11 yet.
*
*/

/* The following prevents "kernel_version" from being set in this file. */
#define __NO_VERSION__

/* PCMCIA headers generated during PCMCIA package installation */
#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>

/* Module related headers, non-module drivers should not include */
#include <linux/version.h>

/* Standard driver includes */
#include <assert.h>
#include <linux/delay.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/kernel_stat.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <linux/errno.h>

/* Ethernet and network includes */
#include <linux/if_ether.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/if_arp.h>
#include <linux/ioport.h>

/* Card and Driver services includes */
#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ds.h>

/* Local Includes */
#include <wlan/version.h>
#include <wlan/wlan_compat.h>
#include <wlan/p80211hdr.h>
#include <wlan/p80211mgmt.h>
#include <wlan/am930mib.h>
#include "am930debug.h"
#include "am930di.h"
#include "am930mac.h"
#include "am930hw.h"
#include "am930mgr.h"
#include "am930llc.h"

#define FILEBIT	DBMAC

/*================================================================*/
/* Local Functions  */

static void am930mac_rxmgmt( am930mac_t *mac, am930rxfrm_t *frm);
static void am930mac_rxdata( am930mac_t *mac, am930rxfrm_t *frm);

/*----------------------------------------------------------------
*	am930mac_construct
*
*	Allocates and initializes the mac object and it's subobjects.
*
*	returns: NULL on failure, the address of the new object
*            otherwise.
----------------------------------------------------------------*/
am930mac_t* am930mac_construct(dev_link_t* di, UINT32 membase)
{
	am930mac_t*					mac;

	DBFENTER;

	/* Allocate the memory */
	mac = kmalloc( sizeof(am930mac_t), GFP_KERNEL);
	if ( mac == NULL )
	{
		DBPRT( DBINIT, "mac object allocation failed!\n");
	}
	else
	{
		/* zero the mac object */
		memset( mac, 0, sizeof(am930mac_t));

		/* save a reference to the di */
		mac->di = di;

		/* set the di priv ptr to mac */
		di->priv = mac;

		/* Create the hw instance */
		mac->hw = am930hw_construct( di->irq.AssignedIRQ,
					di->io.BasePort1, membase, mac) ;

		if ( mac->hw  == NULL)
		{
			DBPRT( DBINIT, "hw object allocation failed!\n");
			am930mac_destruct(mac);
			mac = NULL;
			di->priv = NULL;
		}
		else
		{
			/* Now, call the second part of the hw initialization */
			if ( am930hw_init_rx_tx( mac->hw ) != 0 )
			{
				DBPRT( DBINIT, "hw object rx tx init failed!\n");
				am930mac_destruct(mac);
				mac = NULL; 
				di->priv = NULL;
			}
			else
			{
				mac->mgr = am930mgr_construct( mac, mac->hw );
				if ( mac->mgr != NULL )
				{
					mac->llc = am930llc_construct( mac, mac->mgr);
					if ( mac->llc == NULL )
					{
						DBPRT( DBINIT, "llc object construction failed!\n");
						am930mac_destruct(mac);
						mac = NULL; 
						di->priv = NULL;
					}
				}
				else
				{
					DBPRT( DBINIT, "mgr object construction failed!\n");
					am930mac_destruct(mac);
					mac = NULL; 
					di->priv = NULL;
				}
			}
		}
	}

	DBFEXIT;
	return mac;
}


/*----------------------------------------------------------------
*	am930mac_destruct
*
*	Destroys the subobjects and the mac object. Caller needs to be
*   aware that the pointer passed in is invalid after this call.
*
*	returns: nothing
----------------------------------------------------------------*/
void am930mac_destruct( am930mac_t* mac)
{
	DBFENTER;

	if ( mac->hw != NULL )
	{
		am930hw_destruct( mac->hw );
		mac->hw = NULL;
	}

	if ( mac->llc != NULL )
	{
		am930llc_destruct( mac->llc);
		mac->llc = NULL;
	}

	/* TODO: handle destruction cleanly, call subobject destructors
	         after testing subobject pointers for NULL */

	/* all subobjects handled, now free the mac object */
	kfree_s( mac, sizeof(am930mac_t));

	DBFEXIT;
	return;
}


/*----------------------------------------------------------------
*	am930mac_ISR
*
*	ISR called from the di ISR. I know this is kindof a waste, but
*	I didn't want the di to know anything about the hw interface.
*
*	returns: nothing
----------------------------------------------------------------*/
void am930mac_ISR( am930mac_t *mac )
{
	DBFENTER;
	if ( mac->hw != NULL )
	{
		am930hw_ISR( mac->hw );
	}
	DBFEXIT;
}


/*----------------------------------------------------------------
*	am930mac_ontxcomplete
*
*	Callback from hw when tx is signaled complete.
*
*	returns: nothing
----------------------------------------------------------------*/
void am930mac_ontxcomplete( am930mac_t *mac, UINT32 txresult )
{
	DBFENTER;

	am930llc_ontxcomplete( mac->llc, txresult);

	DBFEXIT;
	return;
}

/*----------------------------------------------------------------
*	am930mac_rxdata
*
*	Private helper for the rxframe method.
*
*	returns: nothing
----------------------------------------------------------------*/
typedef struct snaphdr_t
{
  unsigned char   dsap;
  unsigned char   ssap;
  unsigned char   ctrl;
  unsigned char   org[3];
  unsigned char   ethertype[2];
} snaphdr_t;
#define SNAP_ID       0x0003aaaa
#define BRIDGE_ENCAP  0xf80000
#define RFC1042_ENCAP 0

void am930mac_rxdata( am930mac_t *mac, am930rxfrm_t *frm)
{
	UINT8				*f;
	UINT32				len;

	DBFENTER;

#ifdef DEBUG_RXHDR_DATA
wlan_dbprt_hdr((UINT8*) frm->data); 
#endif

        if (am930_rfc1042) {
	  unsigned char srcaddr[ETH_ALEN];
	  unsigned char destaddr[ETH_ALEN];
	  p80211_hdr_t *pmac = (p80211_hdr_t *)frm->data;
	  struct ethhdr *peth;
	  snaphdr_t *psnap = (snaphdr_t *)(frm->data + WLAN_HDR_A3_LEN);
	  int delta;
	  unsigned short type = *(unsigned short *)psnap->ethertype;
	  unsigned int xsap = *(unsigned int *)psnap & 0x00ffffff;
	  unsigned int org = (*(unsigned int *)psnap->org) & 0x00ffffff;

	  memcpy(destaddr, pmac->a3.a1, ETH_ALEN);
	  memcpy(srcaddr, pmac->a3.a2, ETH_ALEN);

	  if (xsap != SNAP_ID) {
	    delta = WLAN_HDR_A3_LEN - ETH_HLEN;
	    peth = (struct ethhdr *)(frm->data + delta);
	    peth->h_proto = htons(frm->len - WLAN_HDR_A3_LEN - WLAN_CRC_LEN);
	  } else {
	    if (org == BRIDGE_ENCAP) {
	      delta = WLAN_HDR_A3_LEN + sizeof(struct snaphdr_t) - ETH_HLEN;
	      peth = (struct ethhdr *)(frm->data + delta);
	      peth->h_proto = type;
	    } else {
	      if (org == RFC1042_ENCAP) {
		delta = WLAN_HDR_A3_LEN
		  + sizeof(struct snaphdr_t) - ETH_HLEN;
		peth = (struct ethhdr *)(frm->data + delta);
		peth->h_proto = type;
	      } else {
                delta = WLAN_HDR_A3_LEN - ETH_HLEN;
                peth = (struct ethhdr *)(frm->data + delta);
                peth->h_proto = type;
	      }
	    }
	  }
	  f = (UINT8*)frm->data + delta;/* get ptr to eth part */

	  len =	frm->len -		/* len of 802.11 frame rcvd */
	    WLAN_CRC_LEN -		/* len of CRC at end of frame */
	    delta;                      /* len of ethernet header */
	  memcpy(peth->h_dest, destaddr, ETH_ALEN);
	  memcpy(peth->h_source, srcaddr, ETH_ALEN);
	} else {
	  /* strip the 802.11 header and pass it up to the llc layer */
	  f = (UINT8*)frm->data + WLAN_HDR_A3_LEN;/* get ptr to eth part */

	  len =	frm->len -		/* len of 802.11 frame rcvd */
	    WLAN_HDR_A3_LEN -		/* len of 802.11 (3 addrs's) header */
	    WLAN_CRC_LEN;		/* len of CRC at end of frame */
	}
	am930llc_rxframe( mac->llc, f, len);

	kfree_s( frm, sizeof(am930rxfrm_t) + frm->len);
	DBFEXIT;
	return;
}


/*----------------------------------------------------------------
*	am930mac_rxframe
*
*	Public method called when a frame has been received at a
*   lower layer and needs to be handled by this layer.
*
*	returns: nothing
----------------------------------------------------------------*/
void am930mac_rxframe( am930mac_t *mac, am930rxfrm_t *frm)
{
	p80211_hdr_t	*hdr;
	DBFENTER;

	hdr = (p80211_hdr_t*)frm->data;

	WLAN_ASSERT(WLAN_GET_FC_MOREFRAG(ieee2host16(hdr->a3.fc)) == 0);

	switch( WLAN_GET_FC_FTYPE(ieee2host16(hdr->a3.fc)) )
	{
		case WLAN_FTYPE_MGMT:
			am930mac_rxmgmt( mac, frm);
			break;
		case WLAN_FTYPE_DATA:
			am930mac_rxdata( mac, frm);
			break;
		case WLAN_FTYPE_CTL:
		default:
			WLAN_LOG_WARNING0("Ctl or unknown frame type received\n"); 
	}

	DBFEXIT;
	return;
}


/*----------------------------------------------------------------
*	am930mac_rxframe_err
*
*	returns: nothing
----------------------------------------------------------------*/
void am930mac_rxframe_err(am930mac_t *mac)
{
	am930llc_rxframe_err(mac->llc);
	return;
}


/*----------------------------------------------------------------
*	am930mac_rxmgmt
*
*	Private helper for the rxframe method.
*
*	returns: nothing
----------------------------------------------------------------*/
void am930mac_rxmgmt( am930mac_t *mac, am930rxfrm_t *frm)
{
	DBFENTER;

/* lets have a look at the frame */
#ifdef DEBUG_RXHDR_MGMT
wlan_dbprt_hdr((UINT8*) frm->data);
#endif

	am930mgr_rxmgmt( mac->mgr, frm);

	DBFEXIT;
	return;
}


/*----------------------------------------------------------------
*	am930mac_mk_rxfrmbuf
*
*	Public method called from the hw object to allocate the space
*	for the frame data when an rx int has been received. 
*	NOTE: this function may go away later, it's here now in case
*         we need some special handling. This function is called
*		  in an interrupt context!
*
*	returns: the address of the rxfrmbuf structure allocated or
*			 NULL on error.
----------------------------------------------------------------*/
am930rxfrm_t *am930mac_mk_rxfrmbuf( am930mac_t *mac, UINT32 len )
{
	am930rxfrm_t	*frm;

	DBFENTER;
	frm = kmalloc( sizeof(am930rxfrm_t) + len, GFP_ATOMIC);
	if ( frm != NULL )
	{
		memset( frm, 0, sizeof(am930rxfrm_t) + len);
	}
	DBFEXIT;
	return frm;
}


static unsigned char eth2_llc[] = {0xaa, 0xaa, 3, 0, 0, 0};
/*----------------------------------------------------------------
*	am930mac_txdata
*
*	Public method called from the llc object to send a frame. The
*	frame has already been formatted for tx. The 802.11 header is
*	already in place.
*
*	returns: nothing
----------------------------------------------------------------*/
UINT32 am930mac_txdata( am930mac_t *mac, UINT8* daddr, UINT8 *saddr,
						UINT8 *frm, UINT32 len)
{
	UINT8			*d;
	UINT32			len802_11;
	UINT32			result = 0;
	p80211_hdr_t	*hdr;
	struct ethhdr *eh = (struct ethhdr *)frm;
	int translate = 0;
	DBFENTER;

	if ((ntohs(eh->h_proto) == ETH_P_IP || ntohs(eh->h_proto) == 
	    ETH_P_RARP || ntohs(eh->h_proto) == ETH_P_ARP) && 
	    am930_rfc1042)
	  translate = 1;

	if (translate) 
	  len802_11 = len + WLAN_HDR_A3_LEN - ETH_HLEN + 
	    sizeof(struct snaphdr_t);
	else
	/* Calculate the length of the 802.11 frame for encapsulation*/
	len802_11 = len + WLAN_HDR_A3_LEN; 

	/* alloc the 802.11 buffer */
	d = kmalloc( len802_11, GFP_ATOMIC); 

	if ( d == NULL )
	{
		DBPRT( DBTXPATH, "Memory low, frame dropped\n");
	}
	else
	{
		/* clear the buffer */
		memset( d, 0, len802_11);	

		hdr = (p80211_hdr_t*)d;

		if (translate) {
		  memcpy( d + WLAN_HDR_A3_LEN + sizeof(struct snaphdr_t), 
			  frm + ETH_HLEN, len - ETH_HLEN);
		  memcpy( d + WLAN_HDR_A3_LEN, eth2_llc, sizeof(eth2_llc));
		  memcpy( d + WLAN_HDR_A3_LEN + sizeof(eth2_llc), 
			  &(eh->h_proto), 2);
		} else
		/* copy the ethernet frame to the new buffer */
		memcpy( d + WLAN_HDR_A3_LEN, frm, len);

		switch( mac->wlan_mode )
		{
			case WLAN_MODE_NOTJOINED:
				WLAN_LOG_WARNING0("tx attempt, not joined, frame dropped\n");
				break;

			case WLAN_MODE_ADHOC:
				/* copy the addresses to the 802.11 header */
				memcpy( hdr->a3.a1, daddr, WLAN_ADDR_LEN);
				memcpy( hdr->a3.a2, saddr, WLAN_ADDR_LEN);
				memcpy( hdr->a3.a3, mac->mgr->curr_bssid, WLAN_BSSID_LEN);
	
				/* set the rest of the 802.11 header */
				hdr->a3.fc = host2ieee16( 
				WLAN_SET_FC_FTYPE(WLAN_FTYPE_DATA ) |
				WLAN_SET_FC_FSTYPE(WLAN_FSTYPE_DATAONLY) );
				result = am930hw_txdata( mac->hw, (UINT8*)d, len802_11, mac->mgr->curr_rate);
				break;
			case WLAN_MODE_INFRA:
				/* copy the addresses to the 802.11 header */
				memcpy( hdr->a3.a1, mac->mgr->curr_bssid, WLAN_BSSID_LEN);
				memcpy( hdr->a3.a2, saddr, WLAN_ADDR_LEN);
				memcpy( hdr->a3.a3, daddr, WLAN_ADDR_LEN);
	
				/* set the rest of the 802.11 header */
				hdr->a3.fc = host2ieee16( 
					WLAN_SET_FC_FTYPE(WLAN_FTYPE_DATA ) |
					WLAN_SET_FC_FSTYPE(WLAN_FSTYPE_DATAONLY) |
					WLAN_SET_FC_TODS(1) );

				result = am930hw_txdata( mac->hw, (UINT8*)d, len802_11, mac->mgr->curr_rate);
				break;
		}
	
		kfree_s( d, len802_11 );
	}
	DBFEXIT;
	return result;
}
