/*
*
*   Linux WLAN 
*
*   The contents of this file are subject to the Mozilla Public
*   License Version 1.0 (the "License"); you may not use this file
*   except in compliance with the License. You may obtain a copy of
*   the License at http://www.mozilla.org/MPL/
*
*   Software distributed under the License is distributed on an "AS
*   IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
*   implied. See the License for the specific language governing
*   rights and limitations under the License.
*
*   The initial developer of the original code is Mark S. Mathews
*   <mark@absoval.com>.  Portions created by Mark S. Mathews
*   are Copyright (C) 1998 AbsoluteValue Software, Inc.  All Rights Reserved.
*   
*	--------------------------------------------------------------------
*
*	The author may be reached as mark@absoval.com, or C/O AbsoluteValue
*	Software Inc., P.O. Box 941149, Maitland, FL, 32794-1149
*
*	Thanks to David Hinds, Donald Becker, and the rest of the Linux
*	developers worldwide for making all of this possible.
*/


typedef struct am930mem
{
	atomic_t	lock;
	UINT32		bits;
	UINT32		start;
	UINT32		len;
	UINT32		fragsize;
} am930mem_t;

void am930mem_init( am930mem_t *mem, UINT32 start, UINT32 len);

UINT32 am930mem_alloc( am930mem_t *mem, UINT32 size );

UINT32 am930mem_free( am930mem_t *mem, UINT32 p, UINT32 len);

UINT32 am930mem_fragsize(void);



/*----------------------------------------------------------------
*	am930mem_init
*
*	Initializes a structure to handle _very_ simple memory 
*	management.  The structure contains a bitarray that's used
*	to indicate the alloc/free state of each fragment of a block
*	of memory.  The region of memory is identified by the start and
*	len members.  This implementation is used to manage the
*	tx buffer memory on the PRISM card.
*
*	Note: If len is not evenly divisible by sizeof(mem->bits)*8, then
*			there will be a small chunk of memory at the end of the
*			managed region that will never be allocated.
*
*	returns: nothing
----------------------------------------------------------------*/
void am930mem_init( am930mem_t *mem, UINT32 start, UINT32 len)
{
	memset( mem, 0, sizeof(*mem));
	mem->start = start;
	mem->len = len;
	fragsize = len / (sizeof(bits) * 8);
}


/*----------------------------------------------------------------
*	am930mem_alloc
*
*	Tests the structure lock.  If set, returns NULL
*	Searches the bitarray for the appropriate sized chunk,
*	If found
*		sets the bits
*		unlocks the structure
*		returns the offset
*	else
*		unlocks the structure
*		returns NULL
*
*	Note: LSb in mem->bits corresponds to fragment at address mem->start
*	Note: mem->start can't be zero
*	Note: len of alloc'd region isn't saved, caller _must_ save it!
*
*	returns: 0 on failure, non-zero on success
----------------------------------------------------------------*/
UINT32 am930mem_alloc( am930mem_t *mem, UINT32 size )
{
	UINT32	amask;
	UINT32	nbits;
	INT		i;
	UINT32	allocaddr = 0;

	if ( set_bit(0, &mem->lock) == 0)
	{
		nbits = size / mem->fragsize;
		nbits += ((size % mem->fragsize) ? 1 : 0);
		amask = ((UINT32)0xffffffff) >> ((sizeof(mem->bits) * 8) - nbits);

		allocaddr = 0;
		i = 0;
		while (allocaddr == 0 && i < ((sizeof(mem->bits) * 8) - nbits) )
		{
			if ( (amask & mem->bits) == 0 )
			{
				mem->bits |= amask;
				allocaddr = mem->start + (mem->fragsize * i);
			}
			amask = amask << 1;
			i++;
		}
		clear_bit(0, &mem->lock);
	}
	return allocaddr;
}


/*----------------------------------------------------------------
*	am930mem_free
*
*	Tests the structure lock.  If set, returns the value of p
*	clears the bits associated with the fragments in p,
*	returns NULL
*
*	returns: 0 on failure, non-zero on success
----------------------------------------------------------------*/
UINT32 am930mem_free( am930mem_t *mem, UINT32 p, UINT32 size)
{
	UINT32	amask;
	UINT32	nbits;
	INT		i;
	UINT32	allocaddr = 0;

	if ( set_bit(0, &mem->lock) == 0)
	{
		nbits = size / mem->fragsize;
		nbits += ((size % mem->fragsize) ? 1 : 0);
		amask = ((UINT32)0xffffffff) >> ((sizeof(mem->bits)*8) - nbits);

		lshift = p / mem->fragsize;
		amask = amask << lshift;

		mem->bits &= ~amask;

		clear_bit(0, &mem->lock);
	}
	return p;
}
