/* am930mgr.c: Handles the 802.11 mac management functions.
*	--------------------------------------------------------------------
*
*   Linux WLAN 
*
*   The contents of this file are subject to the Mozilla Public
*   License Version 1.0 (the "License"); you may not use this file
*   except in compliance with the License. You may obtain a copy of
*   the License at http://www.mozilla.org/MPL/
*
*   Software distributed under the License is distributed on an "AS
*   IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
*   implied. See the License for the specific language governing
*   rights and limitations under the License.
*
*   The initial developer of the original code is Mark S. Mathews
*   <mark@absoval.com>.  Portions created by Mark S. Mathews
*   are Copyright (C) 1998 AbsoluteValue Software, Inc.  All Rights Reserved.
*   
*	--------------------------------------------------------------------
*
*	The author may be reached as mark@absoval.com, or C/O AbsoluteValue
*	Software Inc., P.O. Box 941149, Maitland, FL, 32794-1149
*
*	Thanks to David Hinds, Donald Becker, and the rest of the Linux
*	developers worldwide for making all of this possible.
*
*	--------------------------------------------------------------------
*	
*/


/* The following prevents "kernel_version" from being set in this file. */
#define __NO_VERSION__

/* PCMCIA headers generated during PCMCIA package installation */
#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>

/* Module related headers, non-module drivers should not include */
#include <linux/version.h>

/* Standard driver includes */
#include <assert.h>
#include <linux/delay.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <linux/errno.h>

/* Ethernet and network includes */
#include <linux/if_ether.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/if_arp.h>
#include <linux/ioport.h>

/* Card and Driver services includes */
#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ds.h>
#include <pcmcia/cisreg.h>

#include <wlan/version.h>
#include <wlan/wlan_compat.h>
#include <wlan/am930mib.h>
#include <wlan/p80211hdr.h>
#include <wlan/p80211mgmt.h>
#include "am930debug.h"
#include "am930di.h"
#include "am930mac.h"
#include "am930hw.h"
#include "am930mgr.h"

#define FILEBIT	DBMGR

/* private methods */
void am930mgr_rxproberesp( am930mgr_t *mgr, am930rxfrm_t *frm);
void am930mgr_rxproberequest( am930mgr_t *mgr, am930rxfrm_t *frm);
void am930mgr_rxbeacon( am930mgr_t *mgr, am930rxfrm_t *frm);
void am930mgr_scancomplete( void *argp );

/*----------------------------------------------------------------
*	am930mgr_construct
*
*	Allocates and initializes the mgr object and it's subobjects.
*
*	returns: NULL on failure, the address of the new object
*            otherwise.
----------------------------------------------------------------*/
am930mgr_t* am930mgr_construct(am930mac_t *mac, am930hw_t *hw)
{
	am930mgr_t* mgr = NULL;
	DBFENTER;

	if ( mac != NULL )
	{
		mgr = kmalloc( sizeof(am930mgr_t), GFP_KERNEL);
		if ( mgr != NULL )
		{
			memset( mgr, 0, sizeof(am930mgr_t));
			mgr->mac = mac;
			mgr->hw = hw;
		}
	}
	DBFEXIT;
	return mgr;
}

/*----------------------------------------------------------------
*	am930mgr_destruct
*
*	destroys the mgr object
*
*	returns: nothing
----------------------------------------------------------------*/
void am930mgr_destruct( am930mgr_t *mgr )
{
	knownbss_t	*bss;
	DBFENTER;
	if ( mgr != NULL )
	{
		/* Free the remaining bss's */
		while ( mgr->bsslist != NULL )
		{
			bss = mgr->bsslist;
			mgr->bsslist = bss->next;
			/* TODO: when we add ttl, need to kill of any timer funcs */
			kfree_s( bss, sizeof( knownbss_t )); 
		}
		kfree_s(mgr, sizeof(am930mgr_t));
	}
	DBFEXIT;
}


/*----------------------------------------------------------------
*	am930mgr_assoc_begin_sta
*
*	Start the station association procedure.  Namely, send an
*	association request frame to the AP.
*
*	returns: nothing for now
----------------------------------------------------------------*/
void am930mgr_assoc_begin_sta(am930mgr_t *mgr)
{
	wlan_fr_assocreq_t		*f;
	wlan_ie_ssid_t			*ssid;
	wlan_ie_supp_rates_t	*supp_rates;

	DBFENTER;

	/* initial setup of frame */
	f = kmalloc( sizeof(wlan_fr_assocreq_t) + WLAN_ASSOCREQ_FR_MAXLEN, GFP_ATOMIC );
	if ( f != NULL )
	{
		memset( f, 0, sizeof(wlan_fr_assocreq_t) + WLAN_ASSOCREQ_FR_MAXLEN);
		f->buf = ((UINT8*)f) + sizeof(wlan_fr_assocreq_t);
		f->len = WLAN_ASSOCREQ_FR_MAXLEN;
		wlan_mgmt_encode_assocreq(f);	
	
		/* insert values */
		f->hdr->a3.fc = host2ieee16(
			WLAN_SET_FC_FTYPE(WLAN_FTYPE_MGMT) | 
			WLAN_SET_FC_FSTYPE(WLAN_FSTYPE_ASSOCREQ) |
			WLAN_SET_FC_TODS(1) );
		memcpy( f->hdr->a3.a1, mgr->curr_bssid, WLAN_BSSID_LEN);
		memcpy( f->hdr->a3.a2, mgr->hw->addr, WLAN_ADDR_LEN);
		memcpy( f->hdr->a3.a3, mgr->curr_bssid, WLAN_BSSID_LEN);
	
		*(f->cap_info) = host2ieee16(mgr->curr_cap_info);
		*(f->listen_int) = host2ieee16(0);

		/* info elem items */
		ssid = (wlan_ie_ssid_t*)(f->buf + f->len);
		memcpy( ssid,  mgr->curr_ssid, WLAN_SSID_MAXLEN+2);
		f->len += ssid->len + 2;

		supp_rates = (wlan_ie_supp_rates_t*)(f->buf + f->len);
		supp_rates->eid = WLAN_EID_SUPP_RATES;
		supp_rates->len = 2;
		supp_rates->rates[0] = 0x02; /* 1 Mb/s */
		supp_rates->rates[1] = 0x04; /* 2 Mb/s */
		f->len += supp_rates->len + 2;

		/* send the frame */
		am930hw_txdata( mgr->hw, f->buf, f->len, mgr->curr_rate);
	
		/* the response(s) will be handled in auth_rx */
		kfree_s( f, sizeof(wlan_fr_assocreq_t) + WLAN_ASSOCREQ_FR_MAXLEN);
	}

	DBFEXIT;
	return;
}


/*----------------------------------------------------------------
*	am930mgr_assocresp_rx
*
*	Handle incoming association response frames
*
*	returns: nothing for now
----------------------------------------------------------------*/
void am930mgr_assocresp_rx( am930mgr_t *mgr, am930rxfrm_t *frm)
{
	wlan_fr_assocresp_t	f;
	DBFENTER;

	/* decode the frame */
	f.len = frm->len - WLAN_CRC_LEN;
	f.buf = frm->data;
	wlan_mgmt_decode_assocresp(&f);

	/* save values and set mode */
	if ( ieee2host16(*(f.status)) == WLAN_MGMT_STATUS_SUCCESS )
	{
		mgr->curr_aid = ieee2host16(*(f.aid));
		mgr->mac->wlan_mode = WLAN_MODE_INFRA;
		printk(KERN_NOTICE"802.11 Association Successful, AID=%d.\n", 
			mgr->curr_aid);
	}
	else
	{
		am930mgr_prstatus(mgr, ieee2host16(*(f.status)) );
	}

	DBFEXIT;
	return;
}

/*----------------------------------------------------------------
*	am930mgr_auth_begin_sta
*
*	Start the station authentication procedure.  Namely, send an
*	auth frame to the AP.
*
*	returns: nothing for now
----------------------------------------------------------------*/
void am930mgr_authen_begin_sta(am930mgr_t *mgr)
{
	wlan_fr_authen_t	*f;

	DBFENTER;

	/* initial setup of frame */
	f = kmalloc( sizeof(wlan_fr_authen_t) + WLAN_AUTHEN_FR_MAXLEN, GFP_ATOMIC );
	if ( f != NULL )
	{
		memset( f, 0, sizeof(wlan_fr_authen_t) + WLAN_AUTHEN_FR_MAXLEN);
		f->buf = ((UINT8*)f) + sizeof(wlan_fr_authen_t);
		f->len = WLAN_AUTHEN_FR_MAXLEN;
		wlan_mgmt_encode_authen(f);	
	
		/* insert values */
		f->hdr->a3.fc = host2ieee16(
			WLAN_SET_FC_FTYPE(WLAN_FTYPE_MGMT) | 
			WLAN_SET_FC_FSTYPE(WLAN_FSTYPE_AUTHEN) |
			WLAN_SET_FC_TODS(1) );
		memcpy( f->hdr->a3.a1, mgr->curr_bssid, WLAN_BSSID_LEN);
		memcpy( f->hdr->a3.a2, mgr->hw->addr, WLAN_ADDR_LEN);
		memcpy( f->hdr->a3.a3, mgr->curr_bssid, WLAN_BSSID_LEN);
	
		*(f->auth_alg) = host2ieee16(WLAN_AUTH_ALG_OPENSYSTEM);
		*(f->auth_seq) = host2ieee16(1);
	
		/* send the frame */
		am930hw_txdata( mgr->hw, f->buf, f->len, mgr->curr_rate);
	
		/* the response(s) will be handled in auth_rx */
		kfree_s( f, sizeof(wlan_fr_authen_t) + WLAN_AUTHEN_FR_MAXLEN);
	}

	DBFEXIT;
	return;
}


/*----------------------------------------------------------------
*	am930mgr_authen_rx
*
*	Handle incoming authentication frames
*
*	returns: nothing for now
----------------------------------------------------------------*/
void am930mgr_authen_rx( am930mgr_t *mgr, am930rxfrm_t *frm)
{
	wlan_fr_authen_t	f;
	DBFENTER;

	/* decode the frame */
	f.len = frm->len - WLAN_CRC_LEN;
	f.buf = frm->data;
	wlan_mgmt_decode_authen(&f);

	if ( ieee2host16(*(f.auth_alg)) == WLAN_AUTH_ALG_OPENSYSTEM)
	{
		if ( ieee2host16(*(f.auth_seq)) != 2 )
		{
			WLAN_LOG_WARNING0("rx'd authen seq==1\n");
		}
		else
		{
			if ( ieee2host16(*(f.status)) == WLAN_MGMT_STATUS_SUCCESS )
			{
				mgr->is_authen = 1;
				printk(KERN_NOTICE"802.11 Authentication Successful.\n");

				am930mgr_assoc_begin_sta(mgr);
			}
			else
			{
				am930mgr_prstatus(mgr, ieee2host16(*(f.status)) );
			}
		}
	}
	else
	{
		WLAN_LOG_ERROR0("Shared key authen not supported.\n");
	}

	DBFEXIT;
	return;
}


/*----------------------------------------------------------------
*	am930mgr_createbss
*
*	Instructs the hw object to create a bss using the supplied
*	attributes. Note that this implementation only supports Ad-Hoc
*	BSS creation.
*	Arguments:
*		ch		The DS channel to use (1-12)
*		ssid	The SSID to use, assumed to be a C string no more
*				than 32 characters in length. The ssid will be
*				used as just the chars, the \0 is NOT part of the 
*				ssid!
*		bcnrate	The beacon rate for the BSS.
*
*	returns: zero on success, non-zero on failure
----------------------------------------------------------------*/
UINT32 am930mgr_createbss(	am930mgr_t *mgr, 
							UINT32 ch, 
							char   *ssid,
							UINT32 bcn, 
							UINT32 atimwindow)
{
	UINT32			result = 0;
	UINT48			bssid = {0,0,0,0,0,0};
	wlan_bss_ts_t	ts = {0,0,0,0,0,0,0,0};
	wlan_ie_ssid_t  ie_ssid;

	DBFENTER;
		
	memcpy(&(ie_ssid.ssid), ssid, WLAN_SSID_MAXLEN);
	ie_ssid.len = strlen(ie_ssid.ssid);
	ie_ssid.eid = 0;
	if ( am930hw_joinbss( mgr->hw, 
							ch,		/* channel */
							1, 		/* newBSS */
							bssid, 	/* bssid */
							&ie_ssid, 	/* ssid */
							bcn, 	/* bcn interval */
							ts, 	/* ts (timestamp) */
							0)		/* sutro local time */
			== AM930HW_SUCCESS )
	{
		mgr->curr_ch = ch;
		mgr->curr_rate = 20;
		mgr->state |= MGR_ISJOINED;
		printk(KERN_DEBUG "WLAN BSS created, channel=%ld, rate=%ld SSID=%s\n",
				ch, mgr->curr_rate, ssid);
		am930mgr_mibgetitem( mgr, MGMT_CURRENT_BSSID, 
							mgr->curr_bssid, WLAN_BSSID_LEN);
		memcpy( mgr->curr_bssid, bssid, WLAN_BSSID_LEN );
	}
	DBFEXIT;
	return result;
}


/*----------------------------------------------------------------
*	am930mgr_joinbss
*
*	Instructs the hw object to join a bss using the supplied
*	attributes. The only argument is the bssid, the rest of the
*	attributes are obtained from the known bss list.
*
*	Arguments:
*		bssid	the 48 bit id of the bss to join
*
*	returns: zero on success, 
*				1 if the bss isn't found in the list,
*				2 if the hw sync command fails.
----------------------------------------------------------------*/
UINT32 am930mgr_joinbss( am930mgr_t *mgr, UINT48 bssid )
{
	UINT32		result = 0;
	knownbss_t	*curr;
	UINT8		mibbuf;

	DBFENTER;
	/* first find the bss in the known list */
	for (curr = mgr->bsslist; curr != NULL; curr = curr->next)
	{
		if ( memcmp( bssid, curr->bssid, sizeof(UINT48)) == 0 )
			break;
	}
	if ( curr == NULL )
	{
		result = 1;
	}
	else
	{
		if ( WLAN_GET_MGMT_CAP_INFO_ESS(curr->cap_info) )
		{
			/* infrastructure network */
			mibbuf = 1;		/* localmib.networkmode = 1  */
			am930hw_mibset( mgr->hw, SUMIB_LOCAL, LOC_NETWORK_MODE, 
				MI_SZ(LOC_NETWORK_MODE), &mibbuf );

			if ( am930hw_joinbss( mgr->hw, 
									curr->channel, 0, bssid, &curr->ssid, 		/* ssid */
									curr->bcn_int, curr->bss_timestamp,
									curr->sutro_reftime) == AM930HW_SUCCESS )
			{
				mgr->curr_ch = curr->channel;
				mgr->curr_rate = curr->rate;
				mgr->curr_cap_info = curr->cap_info;
				memcpy( mgr->curr_ssid, &curr->ssid, WLAN_SSID_MAXLEN+2);
				mgr->state |= MGR_ISJOINED;
				printk(KERN_DEBUG 
					"ESS joined, ch=%ld, rate=%ld,"
					"bssid=%02x:%02x:%02x:%02x:%02x:%02x\n",
					curr->channel, curr->rate, 
					bssid[0], bssid[1], bssid[2], bssid[3], bssid[4], bssid[5]);
				memcpy( mgr->curr_bssid, bssid, WLAN_BSSID_LEN );

				am930mgr_authen_begin_sta(mgr);

/*			format and send auth frame (wait for and handle response)
			format and send assoc frame (wait for and handle response)
*/
			}
			else
			{
				result = 2;
			}
		}
		else
		{
			/* adhoc network */
			if ( am930hw_joinbss( mgr->hw, 
									curr->channel, 0, bssid, &curr->ssid, 		/* ssid */
									curr->bcn_int, curr->bss_timestamp,
									curr->sutro_reftime) == AM930HW_SUCCESS )
			{
				mgr->curr_ch = curr->channel;
				mgr->curr_rate = curr->rate;
				mgr->state |= MGR_ISJOINED;
				printk(KERN_DEBUG 
					"IBSS joined, ch=%ld, rate=%ld,"
					"bssid=%02x:%02x:%02x:%02x:%02x:%02x\n",
					curr->channel, curr->rate, 
					bssid[0], bssid[1], bssid[2], bssid[3], bssid[4], bssid[5]);
				memcpy( mgr->curr_bssid, bssid, WLAN_BSSID_LEN );
				mgr->mac->wlan_mode = WLAN_MODE_ADHOC;
			}
			else
			{
				result = 2;
			}
		}
	}
	DBFEXIT;
	return result;
}


/*----------------------------------------------------------------
*	am930mgr_[get|set]mib[item]
*
*	These are pass-through calls used to prevent the llc object
*	from knowing about the hw object.
*
*	returns: zero on success, non-zero on failure (see hw)
----------------------------------------------------------------*/
UINT32 am930mgr_mibget( am930mgr_t *mgr, UINT32 mibsec, UINT32 len, void *buf)
{
	return am930hw_mibget( mgr->hw, mibsec, 0, len, buf);
}

UINT32 am930mgr_mibgetitem( am930mgr_t *mgr, UINT32 itemid, 
							void *buf, UINT32 len)
{
	return am930hw_mibgetitem( mgr->hw, itemid, buf, len);
}

UINT32 am930mgr_mibset( am930mgr_t *mgr, UINT32 mibsec, UINT32 len, void *buf)
{
	return am930hw_mibset( mgr->hw, mibsec, 0, len, buf);
}

UINT32 am930mgr_mibsetitem( am930mgr_t *mgr, UINT32 itemid, 
							void *buf, UINT32 len)
{
	return am930hw_mibsetitem( mgr->hw, itemid, buf, len);
}


/*----------------------------------------------------------------
*	am930mgr_prstatus
*
*	Log a warning message based on the contents of the Status
*	Code field of an 802.11 management frame.  Defines are in
*	p80211mgmt.h and are derived from 802.11-1997 pp 54.
*
*	returns: nothing
----------------------------------------------------------------*/
void am930mgr_prstatus(am930mgr_t *mgr, UINT16 status )
{
	switch( status )
	{
		case WLAN_MGMT_STATUS_UNSPEC_FAILURE:
			WLAN_LOG_ERROR0("Status code == Unspecified error.\n");
			break;
		case WLAN_MGMT_STATUS_CAPS_UNSUPPORTED:
			WLAN_LOG_ERROR0("Status code == Can't support all requested capabilities.\n");
			break;
		case WLAN_MGMT_STATUS_REASSOC_NO_ASSOC:
			WLAN_LOG_ERROR0("Status code == Reassoc denied, can't confirm original Assoc.\n");
			break;
		case WLAN_MGMT_STATUS_ASSOC_DENIED_UNSPEC:
			WLAN_LOG_ERROR0("Status code == Assoc denied, because we don't feel like it.\n");
			break;
		case WLAN_MGMT_STATUS_UNSUPPORTED_AUTHALG:
			WLAN_LOG_ERROR0("Status code == Peer doesn't support authen algorithm.\n");
			break;
		case WLAN_MGMT_STATUS_RX_AUTH_NOSEQ:
			WLAN_LOG_ERROR0("Status code == Authen frame received out of sequence.\n");
			break;
		case WLAN_MGMT_STATUS_CHALLENGE_FAIL:
			WLAN_LOG_ERROR0("Status code == Authen rejected, challenge failure.\n");
			break;
		case WLAN_MGMT_STATUS_AUTH_TIMEOUT:
			WLAN_LOG_ERROR0("Status code == Authen rejected, timeout waiting for next frame.\n");
			break;
		case WLAN_MGMT_STATUS_ASSOC_DENIED_BUSY:
			WLAN_LOG_ERROR0("Status code == Assoc denied, AP too busy.\n");
			break;
		case WLAN_MGMT_STATUS_ASSOC_DENIED_RATES:
			WLAN_LOG_ERROR0("Status code == Assoc denied, we haven't enough basic rates.\n");
			break;
		default:
			WLAN_LOG_ERROR1("Unknown status code %d.\n", status);
			break;
	}
}


/*----------------------------------------------------------------
*	am930mgr_rxbeacon
*
*	Handles beacon frames.
*
*	returns: nothing
----------------------------------------------------------------*/
void am930mgr_rxbeacon( am930mgr_t *mgr, am930rxfrm_t *frm)
{
	DBFENTER;
	DBFEXIT;
	return;
}


/*----------------------------------------------------------------
*	am930mgr_rxmgmt
*
*	Entry point for the reception and handling of 802.11 management
*	frames.	Makes a determination of the frame type and then calls
*	the appropriate function.
*
*	returns: zero on success, non-zero on failure
----------------------------------------------------------------*/
void am930mgr_rxmgmt( am930mgr_t *mgr, am930rxfrm_t *frm)
{
	p80211_hdr_t	*hdr;
	DBFENTER;

	hdr = (p80211_hdr_t*)frm->data;

	switch( WLAN_GET_FC_FSTYPE(ieee2host16(hdr->a3.fc)) )
	{
		case WLAN_FSTYPE_ASSOCREQ:
			DBRXFRMPRT( DBRXPATH | DBRXDATA, "Assoc req", frm); 
			break;

		case WLAN_FSTYPE_ASSOCRESP:
			DBRXFRMPRT( DBRXPATH | DBRXDATA, "Assoc resp", frm); 
			am930mgr_assocresp_rx( mgr, frm);
			break;
		case WLAN_FSTYPE_REASSOCREQ:
			DBRXFRMPRT( DBRXPATH | DBRXDATA, "Reassoc req", frm); 
			break;

		case WLAN_FSTYPE_REASSOCRESP:
			DBRXFRMPRT( DBRXPATH | DBRXDATA, "Reassoc resp", frm); 
			break;

		case WLAN_FSTYPE_PROBEREQ:
			DBRXFRMPRT( DBRXPATH | DBRXDATA, "Probe req", frm); 
			am930mgr_rxproberequest( mgr, frm);
			break;

		case WLAN_FSTYPE_PROBERESP:
			DBRXFRMPRT( DBRXPATH | DBRXDATA, "Probe resp", frm); 
			am930mgr_rxproberesp( mgr, frm);
			break;

		case WLAN_FSTYPE_BEACON:
			DBRXFRMPRT( DBRXPATH | DBRXDATA, "Beacon", frm); 
			am930mgr_rxbeacon( mgr, frm);
			break;

		case WLAN_FSTYPE_ATIM:
			DBRXFRMPRT( DBRXPATH | DBRXDATA, "ATIM", frm); 
			break;

		case WLAN_FSTYPE_DISASSOC:
			DBRXFRMPRT( DBRXPATH | DBRXDATA, "Disassoc", frm); 
			break;

		case WLAN_FSTYPE_AUTHEN:
			DBRXFRMPRT( DBRXPATH | DBRXDATA, "Deauth", frm); 
			am930mgr_authen_rx( mgr, frm);
			break;

		case WLAN_FSTYPE_DEAUTHEN:
			DBRXFRMPRT( DBRXPATH | DBRXDATA, "Deauth", frm); 
			break;

		default:
			DBRXFRMPRT( DBRXPATH | DBRXDATA, "Unknown", frm);
	}

	kfree_s( frm, sizeof(am930rxfrm_t) + frm->len);

	DBFEXIT;
	return;
}


/*----------------------------------------------------------------
*	am930mgr_rxproberesp
*
*	Handles probe response management frames. This function shold
*	only be called when we are in an active scanning state.
*	Otherwise the frame is ignored.
*	When the probe response is received, we check to see if the
*	bss owning the responder is a "known bss". If not, we add this
*	bss to the list.
*
*	returns: nothing
----------------------------------------------------------------*/
void am930mgr_rxproberesp( am930mgr_t *mgr, am930rxfrm_t *frm)
{
	knownbss_t			*bssptr;
	wlan_fr_proberesp_t	f;
	DBFENTER;

	/* decode the frame */
	f.len = frm->len - WLAN_CRC_LEN;
	f.buf = frm->data;
	wlan_mgmt_decode_proberesp(&f);

	/* search for the bss */
	bssptr = mgr->bsslist;
	while ( bssptr != NULL )
	{
		if ( memcmp(bssptr->bssid, f.hdr->a3.a3, WLAN_BSSID_LEN) == 0 )
			break;
		bssptr = bssptr->next;
	}

	if ( bssptr == NULL )
	{
		/* it's not in the list, add it */
		bssptr = kmalloc( sizeof( knownbss_t ), GFP_ATOMIC );
		if ( bssptr != NULL )  /* handle failure by just dropping the Presp. */
		{
			memset( bssptr, 0, sizeof(knownbss_t) );

			/* save the bss info */
			memcpy( bssptr->bssid, f.hdr->a3.a3, WLAN_BSSID_LEN);
			memcpy( bssptr->bss_timestamp, f.ts, sizeof(UINT64)); 
			bssptr->bcn_int = ieee2host16(*f.bcn_int);
			bssptr->cap_info = ieee2host16(*f.cap_info);
			memcpy( &bssptr->ssid, f.ssid, WLAN_SSID_MAXLEN+2);
			bssptr->channel = f.ds_parms->curr_ch;
			bssptr->rate = 20; /*TODO: parse the supp rates IE */
			bssptr->sutro_reftime = amd2host32(frm->local_time); // am930 specific 
			bssptr->ttl = 0;

			bssptr->next = mgr->bsslist;
			mgr->bsslist = bssptr;
		}
	}
	else
	{
		/* the bss is in the list, update ttl */
	}

	DBFEXIT;
	return;
}


/*----------------------------------------------------------------
*	am930mgr_rxproberequest
*
*	Handles probe request management frames.
*
*	returns: nothing
----------------------------------------------------------------*/
void am930mgr_rxproberequest( am930mgr_t *mgr, am930rxfrm_t *frm)
{
	DBFENTER;
	DBFEXIT;
	return;
}


/*----------------------------------------------------------------
*	am930mgr_scan
*
*	Set up scanning managed by the mgr object. This function
*	begins the scanning of a sequence of channels. We call the
*	hw scan method registering the callback so we can setup the
*	scan the of the next channel. We pick up scan results via the
*	rx_mgmt method via probe response and beacon frames.
*	scantype:	0: active	1: passive
*	bsstype:	BIT0: independent	BIT1: infrastructure
*	timech is in Kus
*
*	returns: zero on success, 
*			1 - already scanning
*			2 - scan command failed
----------------------------------------------------------------*/
UINT32 am930mgr_scan( am930mgr_t *mgr, UINT scantype, UINT bsstype,
						UINT48 bssid, UINT startch, UINT endch, UINT timech)
{
	UINT32 result = 0;
	UINT32 cntl = 0;
	knownbss_t	*bss;
	DBFENTER;

	if ( mgr->state & MGR_ISSCANNING )
	{
		DBPRT( DBMGRSTATE, "mgr_scan called when already scanning!\n");
		result = 1;
	}
	else
	{
		/* clear the bsslist  */
		while( mgr->bsslist != NULL)
		{
			bss = mgr->bsslist;
			mgr->bsslist = mgr->bsslist->next;
			kfree_s( bss, sizeof( knownbss_t )); 
		}

		mgr->state |= MGR_ISSCANNING;
		mgr->scantype = scantype;
		mgr->scanch = startch;
		mgr->scanendch = endch;
		mgr->scantimech = timech;
		memcpy( mgr->scanbssid, bssid, sizeof(bssid));
		mgr->scanbsstype = bsstype;
		if ( scantype == 0 )	cntl |= AM930HWSCAN_ACTIVE;
		else 					cntl |= AM930HWSCAN_PASSIVE;
		if ( bsstype & BIT0 )	cntl |= AM930HWSCAN_IBSS;
		if ( bsstype & BIT1 )	cntl |= AM930HWSCAN_ESS;
		if ( am930hw_scan( mgr->hw, cntl, bssid, startch, timech, 
				am930mgr_scancomplete, mgr ) == AM930HW_FAILURE )
		{
			DBPRT( DBFWCMD, "Initial scan command failed!\n");
			result = 2;
			mgr->state &= ~MGR_ISSCANNING;
		}
		DBPRT( DBMGRCMD, "scanning channel %d\n", mgr->scanch);

		/* Now the mgr object just waits for mgmt frames and the scan
			complete call to signal the completion of the scan of each
			channel
		*/
	}
	DBFEXIT;
	return result;
}

/*----------------------------------------------------------------
*	am930mgr_scancomplete
*
*	Private method that handles the scan complete interrupt (when
*	scanning is begun by mgr). It checks to see if we are finished 
*	with all of the channels to be scanned. If so, then we set the
*	mgr object to a non-scanning state. If not, then we send a new
*	scan command to the hw.
*
*	returns: nothing
----------------------------------------------------------------*/
void am930mgr_scancomplete( void *argp )
{
	am930mgr_t	*mgr = (am930mgr_t *)argp;
	UINT32 cntl = 0;
	DBFENTER;

	if ( (mgr->state & MGR_ISSCANNING) == 0)
	{
		DBPRT( DBMGRSTATE, "scancomplet called when not in scanning state!\n");
	}
	else
	{
		if ( mgr->scanch >= mgr->scanendch )
		{
			/* we're done, clear scanning state */
			mgr->state &= ~MGR_ISSCANNING;
		}
		else
		{
			/* set up for the next channel */
			if ( mgr->scantype == 0 )	cntl |= AM930HWSCAN_ACTIVE;
			else 						cntl |= AM930HWSCAN_PASSIVE;
			if ( mgr->scanbsstype & BIT0 )	cntl |= AM930HWSCAN_IBSS;
			if ( mgr->scanbsstype & BIT1 )	cntl |= AM930HWSCAN_ESS;

			mgr->scanch++;

			if ( am930hw_scan( mgr->hw, cntl, mgr->scanbssid, mgr->scanch, 
						mgr->scantimech, am930mgr_scancomplete, mgr ) 
				== AM930HW_FAILURE )
			{
				DBPRT( DBFWCMD, "subsequent scan command failed!\n");
				mgr->state &= ~MGR_ISSCANNING;
			}
			DBPRT( DBMGRCMD, "scanning channel %d\n", mgr->scanch);
		}
	}
	DBFEXIT;
	return;
}


